@@ -35,6 +35,7 @@ CIM data type                             Python type
 ========================================  =====================================
 boolean                                   :class:`py:bool`
 char16                                    :term:`string`
+                                          or :class:`~pywbem.Char16`
 string                                    :term:`string`
 string (EmbeddedInstance)                 :class:`~pywbem.CIMInstance`
 string (EmbeddedObject)                   :class:`~pywbem.CIMInstance`
@@ -73,15 +74,17 @@ objects, :term:`string`, in addition to
 
 from __future__ import absolute_import
 
-from datetime import tzinfo, datetime, timedelta
-import re
-import warnings
 import copy
-import traceback
-import six
+import re
 
-from .config import DEBUG_WARNING_ORIGIN, ENFORCE_INTEGER_RANGE
-from ._utils import _ensure_unicode, _hash_item, _format, _to_unicode
+from datetime import tzinfo, datetime, timedelta
+import six
+from .config import ENFORCE_INTEGER_RANGE
+from ._utils import _ensure_unicode, _hash_item, _format, _to_unicode, \
+    _eq_item
+
+if six.PY3:
+    from datetime import timezone
 
 if six.PY2:
     # pylint: disable=invalid-name,undefined-variable
@@ -94,93 +97,68 @@ else:
 __all__ = ['cimtype', 'type_from_name', 'MinutesFromUTC', 'CIMType',
            'CIMDateTime', 'CIMInt', 'Uint8', 'Sint8', 'Uint16', 'Sint16',
            'Uint32', 'Sint32', 'Uint64', 'Sint64', 'CIMFloat', 'Real32',
-           'Real64']
+           'Real64', 'Char16', '_CIMComparisonMixin']
 
 
 class _CIMComparisonMixin(object):  # pylint: disable=too-few-public-methods
-    """Mixin class providing default implementations for rich comparison
-    operators.
+    """
+    Mixin class providing default implementations for equality test
+    operators and hash function.
+
+    The implementations of ordering tests are also provided and raise an
+    exception because ordering of CIM objects is not supported.
 
     In Python 2, the rich comparison operators (e.g. `__eq__()`) have
-    precedence over the traditional comparator method (`_cmp__()`).
-    In Python 3, the comparator method (`_cmp__()`) no longer exists.
+    precedence over the traditional comparator method (`__cmp__()`).
+    In Python 3, the comparator method (`__cmp__()`) no longer exists.
     Therefore, implementing the rich comparison operators works in both.
-
-    The default implementations delegate to a comparator method `_cmp()`
-    implemented by subclasses. This requires that the subclasses can
-    define total ordering. (If they cannot, this mixin class cannot be
-    used).
     """
 
+    __slots__ = []
+
     def __eq__(self, other):
         """
-        Invoked when two CIM objects are compared with the `==` operator.
+        Equality test for two CIM objects.
 
-        The comparison is delegated to the `_cmp()` method.
+        The comparison must be implemented in the derived class.
         """
-        return self._cmp(other) == 0
+        raise NotImplementedError
 
     def __ne__(self, other):
         """
-        Invoked when two CIM objects are compared with the `!=` operator.
+        Non-equality test for two CIM objects.
 
-        The comparison is delegated to the `_cmp()` method.
+        The comparison is delegated to the `__eq__()` method, because we
+        require it to be implemented.
+        See https://stackoverflow.com/a/30676267/1424462 for discussion about
+        delegating to == vs. __eq__().
         """
-        return self._cmp(other) != 0
+        return not self.__eq__(other)
 
-    def __ordering_deprecated(self):
-        """Deprecated warning for pywbem CIM Objects"""
-        msg = _format("Ordering comparisons involving {0} objects are "
-                      "deprecated.", self.__class__.__name__)
-        if DEBUG_WARNING_ORIGIN:
-            msg += "\nTraceback:\n" + ''.join(traceback.format_stack())
-        warnings.warn(msg, DeprecationWarning, stacklevel=3)
+    def __raise_ordering_not_supported(self, other, op):
+        """
+        Function to raise a TypeError indicating that ordering of this class
+        is not supported.
+        """
+        raise TypeError(
+            "'{}' not supported between instances of '{}' and '{}'".
+            format(op, type(self), type(other)))
 
     def __lt__(self, other):
-        """
-        Invoked when two CIM objects are compared with the `<` operator.
-
-        The comparison is delegated to the `_cmp()` method.
-        """
-        self.__ordering_deprecated()
-        return self._cmp(other) < 0
+        ""  # Avoids docstring to be inherited
+        self.__raise_ordering_not_supported(other, '<')
 
     def __gt__(self, other):
-        """
-        Invoked when two CIM objects are compared with the `>` operator.
-
-        The comparison is delegated to the `_cmp()` method.
-        """
-        self.__ordering_deprecated()
-        return self._cmp(other) > 0
-
-    def __le__(self, other):
-        """
-        Invoked when two CIM objects are compared with the `<=` operator.
-
-        The comparison is delegated to the `_cmp()` method.
-        """
-        self.__ordering_deprecated()
-        return self._cmp(other) <= 0
+        ""  # Avoids docstring to be inherited
+        self.__raise_ordering_not_supported(other, '>')
 
     def __ge__(self, other):
-        """
-        Invoked when two CIM objects are compared with the `>=` operator.
+        ""  # Avoids docstring to be inherited
+        self.__raise_ordering_not_supported(other, '>=')
 
-        The comparison is delegated to the `_cmp()` method.
-        """
-        self.__ordering_deprecated()
-        return self._cmp(other) >= 0
-
-    def _cmp(self, other):
-        """
-        Interface definition for comparator method to be provided by
-        subclasses, as follows:
-        * If self == other, 0 must be returned.
-        * If self < other, -1 must be returned.
-        * If self > other, +1 must be returned.
-        """
-        raise NotImplementedError
+    def __le__(self, other):
+        ""  # Avoids docstring to be inherited
+        self.__raise_ordering_not_supported(other, '<=')
 
     def __hash__(self):
         """
@@ -197,7 +175,71 @@ class _CIMComparisonMixin(object):  # pylint: disable=too-few-public-methods
         raise NotImplementedError
 
 
-class MinutesFromUTC(tzinfo):
+class SlottedPickleMixin(object):
+    """
+    On Python 2, the built-in 'pickle' module uses pickle protocol 0 by default.
+    Using protocol 0 causes pickle to raise TypeError for objects with slots
+    that do not define methods __getstate__() and __setstate__().
+
+    In Python 3.0, the default pickle protocol changed to protocol 3 and in
+    Python 3.8 to protocol 4, both of which support objects with slots by
+    default, without requiring that they define methods __getstate__() and
+    __setstate__().
+
+    That led to the idea to provide these methods only when running on
+    Python 2. However, it turns out that objects pickled before this change
+    (using the built-in methods) save all attributes in one shot by saving
+    their __dict__ value, so the default unpickling after introduction of
+    slots attempts to set the __dict__ which fails with AttributeError.
+    Therefore, these methods need to be provided also for Python 3.
+
+    Also, because code in pywbem_mock/_resolvermixin.py dynamically added
+    the non-existing attributes 'classorigin' and 'propagated' to CIMClass
+    objects (that code has been removed meanwhile), the unpickling may
+    encounter CIMClass objects that still have these attributes set. That
+    is handled by ignoring these attributes during unpickling.
+
+    These methods support classes with both slotted and non-slotted attributes.
+    Support for non-slottted attributes is needed for example when users
+    define classes derived from CIM object or CIM type classes.
+    """
+
+    __slots__ = []
+
+    def __getstate__(self):
+        ""  # Avoids docstring to be inherited
+
+        dct = {}
+        for attr in self.__slots__:
+            dct[attr] = getattr(self, attr)
+
+        # Support for objects that also have __dict__, e.g. user defined
+        # derived classes that did not define __slots__:
+        if hasattr(self, '__dict__'):
+            for attr in self.__dict__:
+                dct[attr] = getattr(self, attr)
+
+        return dct
+
+    def __setstate__(self, dct):
+        ""  # Avoids docstring to be inherited
+
+        # Defer import due to circular import dependencies:
+        # pylint: disable=import-outside-toplevel
+        from pywbem._cim_obj import CIMClass as class_type
+
+        for attr in dct:
+
+            # Compatibility support for CIMClass objects that were pickled
+            # before pywbem 1.2.0:
+            if isinstance(self, class_type) and \
+                    attr in ('classorigin', 'propagated'):
+                continue
+
+            setattr(self, attr, dct[attr])
+
+
+class MinutesFromUTC(SlottedPickleMixin, tzinfo):
     """
     Timezone information (an implementation of :class:`py:datetime.tzinfo`)
     that represents a fixed offset in +/- minutes from UTC and is thus suitable
@@ -235,6 +277,8 @@ class MinutesFromUTC(tzinfo):
                                                   pywbem.MinutesFromUTC(-300))
     """
 
+    __slots__ = ['_offset']
+
     def __init__(self, offset):  # pylint: disable=super-init-not-called
         """
         Parameters:
@@ -250,6 +294,9 @@ class MinutesFromUTC(tzinfo):
         self._offset = offset
 
     def __repr__(self):
+        """
+        Return a string representation suitable for debugging.
+        """
         return _format(
             "MinutesFromUTC("
             "offset={s._offset!A})",
@@ -306,15 +353,42 @@ class MinutesFromUTC(tzinfo):
         return "{sign}{hh:02d}:{mm:02d}".format(sign=sign, hh=hh, mm=mm)
 
 
-class CIMType(object):  # pylint: disable=too-few-public-methods
+class CIMType(SlottedPickleMixin):  # pylint: disable=too-few-public-methods
     """Base type for all CIM data types defined in this package."""
 
+    __slots__ = []
+
     #: The name of the CIM datatype, as a :term:`string`. See
     #: :ref:`CIM data types` for details.
     cimtype = None
 
 
-class CIMDateTime(CIMType, _CIMComparisonMixin):
+class Char16(CIMType, six.text_type):
+    """
+    A value of CIM data type char16.
+
+    This class is derived from :term:`unicode string`.
+
+    Normally, values of CIM data type char16 are represented using
+    :term:`unicode string` objects. This class can be used to represent values
+    of CIM data type char16 when it matters to distinguish them from values of
+    CIM data type string. The only situation where that matters is for
+    keybindings, because that allows properly setting the TYPE attribute on
+    KEYVALUE elements when creating the CIM-XML representation for a keybinding.
+    """
+
+    __slots__ = []
+
+    #: The name of the CIM datatype ``"char16"``
+    cimtype = 'char16'
+
+    # Changing the string content requires using __new__() instead of __init__()
+    def __new__(cls, content=''):
+        ""  # Avoids docstring to be inherited
+        return super(Char16, cls).__new__(cls, _ensure_unicode(content))
+
+
+class CIMDateTime(_CIMComparisonMixin, CIMType):
     """
     A value of CIM data type datetime.
 
@@ -326,6 +400,8 @@ class CIMDateTime(CIMType, _CIMComparisonMixin):
     hash value being based on their public attributes.
     """
 
+    __slots__ = ['__timedelta', '__datetime', '__precision']
+
     #: The name of the CIM datatype ``"datetime"``
     cimtype = 'datetime'
 
@@ -582,11 +658,18 @@ class CIMDateTime(CIMType, _CIMComparisonMixin):
         A positive value indicates minutes east of UTC, and a negative
         value indicates minutes west of UTC.
         """
-        utc = datetime.utcnow()
-        local = datetime.now()
-        if local < utc:
-            return - int(float((utc - local).seconds) / 60 + .5)
-        return int(float((local - utc).seconds) / 60 + .5)
+        if six.PY2:
+            # Effective Python 3.12, utcnow() deprecated, it returns
+            # naive datetime object and warning msg. timezone does
+            # not exist for python 2.7; use utcnow(timezone.utc).
+            utc_unaware = datetime.utcnow()
+        else:
+            utc_unaware = datetime.now(timezone.utc).replace(tzinfo=None)
+
+        local_unaware = datetime.now()
+        if local_unaware < utc_unaware:
+            return - int(float((utc_unaware - local_unaware).seconds) / 60 + .5)
+        return int(float((local_unaware - utc_unaware).seconds) / 60 + .5)
 
     @classmethod
     def now(cls, tzi=None):
@@ -715,22 +798,33 @@ class CIMDateTime(CIMType, _CIMComparisonMixin):
             s=self)
 
     def __getstate__(self):
+        ""  # Avoids docstring to be inherited
         return str(self)
 
     def __setstate__(self, arg):
+        ""  # Avoids docstring to be inherited
         self.__init__(arg)
 
-    def _cmp(self, other):
-        # Defer import due to circular import dependencies:
-        from .cim_obj import cmpitem
+    def __eq__(self, other):
+        """
+        Equality test function for two :class:`~pywbem.CIMDateTime` objects.
+
+        The equality is based on their public attributes.
+
+        Note that a :class:`~pywbem.CIMDateTime` object represents either a
+        time interval or a point in time but never both. Therefore, an object
+        representing an interval is never equal to an object representing a
+        point in time.
+
+        Returns `False` if the `other` object is not a
+        :class:`~pywbem.CIMDateTime` object.
+        """
         if self is other:
-            return 0
-
+            return True
         if not isinstance(other, CIMDateTime):
-            return 1
-
-        return (cmpitem(self.datetime, other.datetime) or
-                cmpitem(self.timedelta, other.timedelta))
+            return False
+        return (_eq_item(self.datetime, other.datetime) and
+                _eq_item(self.timedelta, other.timedelta))
 
     def __hash__(self):
         """
@@ -753,7 +847,7 @@ class CIMDateTime(CIMType, _CIMComparisonMixin):
 class CIMInt(CIMType, _Longint):
     """
     Base type for CIM integer data types. Derived from :class:`~pywbem.CIMType`
-    and :class:`py:int` (for Python 3) or :class:`py:long` (for Python 2).
+    and :class:`py3:int` (for Python 3) or :class:`py2:long` (for Python 2).
 
     This class has a concept of a valid range for the represented integer,
     based upon the capability of the CIM data type as defined in
@@ -795,6 +889,8 @@ class CIMInt(CIMType, _Longint):
         TypeError: int() can't convert non-string with explicit base
     """
 
+    __slots__ = []
+
     #: The minimum valid value for the integer, according to the capabilities
     #: of its CIM data type. See :ref:`CIM data types` for a list of CIM
     #: integer data types.
@@ -806,6 +902,7 @@ class CIMInt(CIMType, _Longint):
     maxvalue = None
 
     def __new__(cls, *args, **kwargs):
+        ""  # Avoids docstring to be inherited
 
         # Python 3.7 removed support for passing the value for int() as a
         # keyword argument named 'x'. It now needs to be passed as a positional
@@ -846,6 +943,9 @@ class Uint8(CIMInt):
 
     For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
     """
+
+    __slots__ = []
+
     #: The name of the CIM datatype
     cimtype = 'uint8'
     #: The minimum valid value for the CIM datatype
@@ -860,6 +960,9 @@ class Sint8(CIMInt):
 
     For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
     """
+
+    __slots__ = []
+
     #: The name of the CIM datatype
     cimtype = 'sint8'
     #: The minimum valid value for the CIM datatype
@@ -874,6 +977,9 @@ class Uint16(CIMInt):
 
     For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
     """
+
+    __slots__ = []
+
     #: The name of the CIM datatype
     cimtype = 'uint16'
     #: The minimum valid value for the CIM datatype
@@ -888,6 +994,9 @@ class Sint16(CIMInt):
 
     For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
     """
+
+    __slots__ = []
+
     #: The name of the CIM datatype
     cimtype = 'sint16'
     #: The minimum valid value for the CIM datatype
@@ -902,6 +1011,9 @@ class Uint32(CIMInt):
 
     For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
     """
+
+    __slots__ = []
+
     #: The name of the CIM datatype
     cimtype = 'uint32'
     #: The minimum valid value for the CIM datatype
@@ -916,6 +1028,9 @@ class Sint32(CIMInt):
 
     For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
     """
+
+    __slots__ = []
+
     #: The name of the CIM datatype
     cimtype = 'sint32'
     #: The minimum valid value for the CIM datatype
@@ -930,6 +1045,9 @@ class Uint64(CIMInt):
 
     For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
     """
+
+    __slots__ = []
+
     #: The name of the CIM datatype
     cimtype = 'uint64'
     #: The minimum valid value for the CIM datatype
@@ -944,6 +1062,9 @@ class Sint64(CIMInt):
 
     For details on CIM integer data types, see :class:`~pywbem.CIMInt`.
     """
+
+    __slots__ = []
+
     #: The name of the CIM datatype
     cimtype = 'sint64'
     #: The minimum valid value for the CIM datatype
@@ -976,6 +1097,8 @@ class CIMFloat(CIMType, float):
     in sets.
     """
 
+    __slots__ = []
+
     # Note: __str__() is added later, for Python 3.
 
     def __repr__(self):
@@ -995,6 +1118,9 @@ class Real32(CIMFloat):
     class, or to use them as dictionary keys or as members in sets. See
     :class:`~pywbem.CIMFloat` for details.
     """
+
+    __slots__ = []
+
     #: The name of the CIM datatype
     cimtype = 'real32'
 
@@ -1007,6 +1133,9 @@ class Real64(CIMFloat):
     class, or to use them as dictionary keys or as members in sets. See
     :class:`~pywbem.CIMFloat` for details.
     """
+
+    __slots__ = []
+
     #: The name of the CIM datatype
     cimtype = 'real64'
 
@@ -1036,12 +1165,12 @@ def cimtype(obj):
     For an array, the type is determined from the first array element
     (CIM arrays must be homogeneous w.r.t. the type of their elements).
     If the array is empty, that is not possible and
-    :exc:`~py:exceptions.ValueError` is raised.
+    :exc:`py:ValueError` is raised.
 
     Note that Python :term:`numbers <number>` are not valid input objects
     because determining their CIM data type (e.g. :class:`~pywbem.Uint8`,
     :class:`~pywbem.Real32`) would require knowing the value range. Therefore,
-    :exc:`~py:exceptions.TypeError` is raised in this case.
+    :exc:`py:TypeError` is raised in this case.
 
     Parameters:
 
@@ -1078,27 +1207,24 @@ def cimtype(obj):
     if isinstance(obj, (datetime, timedelta)):
         return 'datetime'
 
-    try:
-        instancename_type = CIMInstanceName
-    except NameError:
-        # Defer import due to circular import dependencies:
-        from pywbem._cim_obj import CIMInstanceName as instancename_type
+    # Defer import due to circular import dependencies:
+    # pylint: disable=import-outside-toplevel
+    from pywbem._cim_obj import CIMInstanceName as instancename_type
+
     if isinstance(obj, instancename_type):
         return 'reference'
 
-    try:
-        instance_type = CIMInstance
-    except NameError:
-        # Defer import due to circular import dependencies:
-        from pywbem._cim_obj import CIMInstance as instance_type
+    # Defer import due to circular import dependencies:
+    # pylint: disable=import-outside-toplevel
+    from pywbem._cim_obj import CIMInstance as instance_type
+
     if isinstance(obj, instance_type):  # embedded instance
         return 'string'
 
-    try:
-        class_type = CIMClass
-    except NameError:
-        # Defer import due to circular import dependencies:
-        from pywbem._cim_obj import CIMClass as class_type
+    # Defer import due to circular import dependencies:
+    # pylint: disable=import-outside-toplevel
+    from pywbem._cim_obj import CIMClass as class_type
+
     if isinstance(obj, class_type):  # embedded class
         return 'string'
 
@@ -1153,6 +1279,10 @@ def type_from_name(type_name):
       Use `decode()` and `encode()` for strings instead of type conversion
       syntax (in both Python 2 and 3, for consistency).
 
+    * For CIM type "reference", type object :class:`~pywbem.CIMInstanceName`
+      is returned, even though for use in CIM method parameters,
+      :class:`~pywbem.CIMClassName` is also valid.
+
     Parameters:
 
       type_name (:term:`string`):
@@ -1170,6 +1300,7 @@ def type_from_name(type_name):
     """
     if type_name == 'reference':
         # Defer import due to circular import dependencies:
+        # pylint: disable=import-outside-toplevel
         from ._cim_obj import CIMInstanceName
         return CIMInstanceName
     try:
